﻿using Styx;
using Styx.CommonBot;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

#region [Method] - Class Redundancy
using Action = Styx.TreeSharp.Action;
using L = Monkee.Core.Utilities.Log;
using CommonBehaviors.Actions;
using Buddy.Coroutines;
using Styx.CommonBot.Coroutines;
using U = Monkee.Core.Unit;
using SB = Monkee.Core.Helpers.SpellBook;
#endregion

namespace Monkee.Core
{
    static class Spell
    {
        public delegate WoWUnit unitSelectionDelegate(object Context);
        public delegate T Selection<out T>(object Context);
        public static int lastSpellCast;
        private static LocalPlayer Me { get { return StyxWoW.Me; } }
        private static WoWUnit currentTarget { get { return StyxWoW.Me.CurrentTarget; } }
        private static WoWUnit Focused { get { return StyxWoW.Me.FocusedUnit; } }

        #region [Method] - Cast Method
        public static async Task<bool> spellCast(int spell, WoWUnit target)
        {
            // Return false if we can't cast the spell
            if (!SpellManager.CanCast(spell))
                return false;


            // Cast spell, return false if it fails to cast
            if (!SpellManager.Cast(spell, target))
                return false;


            L.combatLog("Cast {0} on {1}", WoWSpell.FromId(spell).Name, target.SafeName);


            // Wait for lag
            await CommonCoroutines.SleepForLagDuration();


            // return true - we've cast the spell successfully.
            return true;
        }


        public static async Task<bool> spellCast(int spell, bool melee = true)
        {
            lastSpellCast = spell;
            return await spellCast(spell, currentTarget == null ? Me : currentTarget);
        }
        
        #endregion

        #region [Method] - Cooldown Detection
        public static double cooldownTimeLeft(this int Spell)
        {
            try
            {
                SpellFindResults Results;
                return SpellManager.FindSpell(Spell, out Results) ? (Results.Override != null ? Results.Override.CooldownTimeLeft.TotalMilliseconds : Results.Original.CooldownTimeLeft.TotalMilliseconds) : 9999;
            }
            catch (Exception xException)
            {
                L.diagnosticLog("Exception in cooldownTimeLeft(): ", xException);
                return 9999;
            }
        }

       
        public static double getGCD() {
            try
            {
                return SpellManager.GlobalCooldownLeft.TotalMilliseconds;
            }
            catch {
                return -1.0;
            }
        } 
        #endregion

        #region Item use
        public static bool CanUseEquippedItem(WoWItem item)
        {
            var itemSpell = Lua.GetReturnVal<string>("return GetItemSpell(" + item.Entry + ")", 0);
            if (string.IsNullOrEmpty(itemSpell))
                return false;

            return item.Usable && item.Cooldown <= 0;
        }

        /// <summary>
        ///  Checks for items in the bag, and returns the first item that has an usable spell from the specified string array.
        /// </summary>
        /// <param name="spellNames"> Array of spell names to be check.</param>
        /// <returns></returns>
        public static WoWItem FindFirstUsableItemByItemId(int itemId)
        {
            List<WoWItem> carried = StyxWoW.Me.CarriedItems;
            // Yes, this is a bit of a hack. But the cost of creating an object each call, is negated by the speed of the Contains from a hash set.
            // So take your optimization bitching elsewhere.
            return (from i in carried
                    let spells = i.Effects
                    where i.ItemInfo != null && spells != null && spells.Count != 0 &&
                          i.Usable &&
                          i.Cooldown == 0 &&
                          i.ItemInfo.RequiredLevel <= StyxWoW.Me.Level &&
                          spells.Any(s => s.Spell != null && i.ItemInfo.Id == itemId)
                    orderby i.ItemInfo.Level descending
                    select i).FirstOrDefault();
        }
        #endregion

        #region Others
        public static double getCastTime(int s) {
            try
            {
                WoWSpell spell = WoWSpell.FromId(s);
                return spell.CastTime;
            }
            catch (Exception ex) {
                L.diagnosticLog("Exception in getCastTime()", ex);
                return 0;
            }
        }

        public static void cancelAura(this int spell) {
            if (Me.HasAura(spell)) {
                WoWAura a = Me.GetAuraById(spell);
                if (a != null && a.Cancellable) {
                    a.TryCancelAura();
                    L.combatLog("Aura {0} canceled!", a.Name); 
                }
            }
        }
        #endregion

        #region [Method] - Spell Charges
        public static uint spellCharges(this int Spell)
        {
            uint spellCharges = Lua.GetReturnVal<uint>("return GetSpellCharges(" + Spell + ")", 0);

            return spellCharges;
        }


        #endregion


        #region [Method] - Ground Casting
        public static async Task<bool> spellCastGround(WoWSpell spell, WoWPoint targetLoc)
        {
            // If we cannot cast the spell, obviously
            // we're not going to take any actions, so just return false
            if (!SpellManager.CanCast(spell))
                return false;

            // If the spell cast fails for whatever other reason
            // then we can't do anything either.
            if (!SpellManager.Cast(spell))
                return false;

            // 'Wait' waits until the condition becomes true or the timeout period has elapsed.
            // False is returned if the waiting condition is false for the entire
            // timeout period (here 1000 milliseconds).
            if (!await Coroutine.Wait(1000, () => StyxWoW.Me.CurrentPendingCursorSpell != null))
            {
                L.diagnosticLog("Cursor didn't turn into the spell!");
                return false;
            }
            SpellManager.ClickRemoteLocation(targetLoc);

            await CommonCoroutines.SleepForLagDuration();
            return true;
        }
        #endregion
    }
}
